/*
   format.c: send a `format unit' command
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/../scsi/scsi.h>
#include <sys/ioctl.h>
#include "var.h"

#define FORMAT_ALLOC_LEN 0xff

/*
   format unit
*/
void format_unit ( char *fname )
{
  int cmdlen, status, addlen, fd, set_timeout, get_timeout;
  struct sg_header *sg;
  char *cdb = bufcdb + SCMD_HEADSIZ;
  char *res = bufres + SCMD_HEADSIZ;
  char *param = cdb + 6;
  char buf[256];

  int fmtdata = 0;
  int cmplist = 0;
  int defect_form = 0;
  int interleave = 0;
  int fov = 1;
  int dpry = 0;
  int dcrt = 0;
  int stpf = 0;
  int ip = 0;
  int dsp = 0;
  int immed = 1;

  fgets ( buf, 256, stdin );
  while ( strncmp ( buf, "===", 3 ) ) {
    if ( !strncmp ( "LUN", buf, 3 ) )
      scsi_lun = atoi ( buf + 3 );
    if ( !strncmp ( "FMTDATA", buf, 7 ) )
      fmtdata = atoi ( buf + 7 );
    else if ( !strncmp ( "CMPLIST", buf, 7 ) )
      cmplist = atoi ( buf + 7 );
    else if ( !strncmp ( "DEFECT", buf, 6 ) )
      defect_form = atoi ( buf + 6 );
    else if ( !strncmp ( "INTERLEAVE", buf, 10 ) )
      interleave = atoi ( buf + 10 );
    else if ( !strncmp ( "FOV", buf, 3 ) )
      fov = atoi ( buf + 3 );
    else if ( !strncmp ( "DPRY", buf, 4 ) )
      dpry = atoi ( buf + 4 );
    else if ( !strncmp ( "DCRT", buf, 4 ) )
      dcrt = atoi ( buf + 4 );
    else if ( !strncmp ( "STPF", buf, 4 ) )
      stpf = atoi ( buf + 4 );
    else if ( !strncmp ( "IP", buf, 2 ) )
      ip = atoi ( buf + 2 );
    else if ( !strncmp ( "DSP", buf, 3 ) )
      dsp = atoi ( buf + 3 );
    else if ( !strncmp ( "IMMED", buf, 5 ) )
      immed = atoi ( buf + 5 );
    fgets ( buf, 256, stdin );
  }

  fd = open ( fname, O_RDWR );

  if ( fd < 0 ) {
    puts ( "*** 001 CAN'T OPEN" );
    fflush ( stdout );
    return;
  }

  set_timeout = 0;
  ioctl ( fd, SG_SET_TIMEOUT, &set_timeout );

  *(cdb +  0) = FORMAT_UNIT;
  *(cdb +  1) = (scsi_lun << 5) |
    ((fmtdata != 0) << 4) | ((cmplist != 0) << 3) | (defect_form & 0x07);
  *(cdb +  2) = 0;
  *(cdb +  3) = u8 ( interleave >> 8 );
  *(cdb +  4) = u8 ( interleave & 0x00ff );
  *(cdb +  5) = 0;

  cmdlen = SCMD_HEADSIZ + 6;
  if ( fmtdata  ) {
    *(param + 0) = 0;
    *(param + 1) = ( ( fov != 0 ) << 7 ) 
      | ( ( dpry != 0 ) << 6 ) 
	| ( ( dcrt != 0 ) << 5 )
	  | ( ( stpf != 0 ) << 4 )
	    | ( ( ip != 0 ) << 3 )
	      | ( ( dsp != 0 ) << 2 )
		| ( ( immed != 0 ) << 1 );
    *(param + 2) = 0;
    *(param + 3) = 0;
    cmdlen += 4;
  }

/*  fprintf ( stderr, "%02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
	   u8(*(cdb + 0)), u8(*(cdb + 1)), u8(*(cdb + 2)), u8(*(cdb + 3)),
	   u8(*(cdb + 4)), u8(*(cdb + 5)), u8(*(cdb + 6)), u8(*(cdb + 7)),
	   u8(*(cdb + 8)), u8(*(cdb + 9)) );
*/
  
  status = scsicmd ( fd, bufres, FORMAT_ALLOC_LEN, bufcdb, cmdlen );
  close ( fd );

/*  fputs ( "Format Unit Return.\n", stderr ); */
  fflush ( stderr );
  if ( status ) {
    sg = (struct sg_header *)bufres;
    puts ( "*** 002 FAIL TO SEND COMMAND" );
    printf ( "set status %d\n", status );
    printf ( "set errcode %d\n", sg->sense_buffer[0] & 0x007f );
    printf ( "set sensekey %d\n", sg->sense_buffer[2] & 0x000f );
    printf ( "set asc %d\n", u8 (sg->sense_buffer[12]) );
    printf ( "set ascq %d\n", u8 (sg->sense_buffer[13]) );
    puts ( "===" );
    fflush ( stdout );
    return;
  }

  puts ( "set status 0" );
  puts ( "===" );
  fflush ( stdout );
}
