/*
   modesense.c: send Mode Sense command
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/../scsi/scsi.h>
#include "var.h"

static int media_direct ( int, int );
static int media_sequential ( int, int );
static int density_sequential ( int, int );
static int media_printer ( int, int );
static int media_optical ( int, int );
static int density_optical ( int, int );
static int media_cdrom ( int, int );
static int density_cdrom ( int, int );

static int error_recover ( char *, char *, int );
static int disconnect_param ( char *, char *, int );
static int format_param ( char *, char *, int );
static int pprinter_interface_param ( char *, char *, int );
static int measure_unit_param ( char *, char *, int );
static int drive_param ( char *, char *, int );
static int sprinter_interface_param ( char *, char *, int );
static int floppy_param ( char *, char *, int );
static int printer_option_param ( char *, char *, int );
static int optical_param ( char *, char *, int );
static int verify_error_recovery_param ( char *, char *, int );
static int cache_control_param ( char *, char *, int );
static int peripheral_param ( char *, char *, int );
static int controlmode_param ( char *, char *, int );
static int mediatype_param ( char *, char *, int );
static int cdrom_param ( char *, char *, int );
static int cdrom_audio_param ( char *, char *, int );



/*
   Mode Sense
*/
void mode_sense ( int pagecode, int pagecontrol, char *fname, int devtype )
{
  int status, addlen, fd;
  struct sg_header *sg;
  char *cdb = bufcdb + SCMD_HEADSIZ;
  char *res = bufres + SCMD_HEADSIZ;
  char buf[256];
  int param_len, block_des, offset, page_len, media_type, device_uniq;
  int density;
  int i, j;

  fd = open ( fname, O_RDWR );
  if ( fd < 0 ) {
    puts ( "*** 001 CAN'T OPEN" );
    fflush ( stdout );
    return;
  }

  scsi_type = devtype;

  *(cdb +  0) = MODE_SENSE;
  *(cdb +  1) = scsi_lun << 5;
  *(cdb +  2) = (pagecontrol <<6 ) | ( (pagecode == 0) ? 0x3f : pagecode );
  *(cdb +  3) = 0;
  *(cdb +  4) = 0xff;
  *(cdb +  5) = 0;

  status = scsicmd ( fd, bufres, SCMD_RECVMAX, bufcdb, SCMD_HEADSIZ + 6 );

  if ( status ) {
    sg = (struct sg_header *)bufres;
    puts ( "*** 002 FAIL TO SEND COMMAND" );
    printf ( "set status %d\n", status );
    printf ( "set errcode %d\n", sg->sense_buffer[0] & 0x007f );
    printf ( "set sensekey %d\n", sg->sense_buffer[2] & 0x000f );
    printf ( "set asc %d\n", u8 (sg->sense_buffer[12]) );
    printf ( "set ascq %d\n", u8 (sg->sense_buffer[13]) );
    puts ( "===" );
    fflush ( stdout );
    return;
  }

  param_len = u8 ( *(res + 0) );
  media_type = u8 ( *(res + 1 ) );
  device_uniq = u8 ( *(res + 2 ) );
  block_des = u8 ( *(res + 3) )>> 3;

  if ( pagecode == 0 ) {
    switch ( scsi_type ){
    case TYPE_DISK:
      media_direct ( media_type, device_uniq ); break;
    case TYPE_TAPE:
      media_sequential ( media_type, device_uniq ); break;
    case TYPE_PRINTER:
      media_printer ( media_type, device_uniq ); break;
    case TYPE_WORM:
    case TYPE_MOD:
      media_optical ( media_type, device_uniq ); break;
    case TYPE_ROM:
      media_cdrom ( media_type, device_uniq ); break;
    default:
      break;
    }
    for ( i = 0; i < block_des; ++i ) {
      density = *(res+4+8*i);
      switch ( scsi_type ) {
      case TYPE_TAPE:
	density_sequential ( i, density ); break;
      case TYPE_WORM:
      case TYPE_MOD:
	density_optical ( i, density ); break;
      case TYPE_ROM:
	density_cdrom ( i, density ); break;
      default:
	break;
      }
      printf ( "set blocknum [concat $blocknum {\"%d\"}]\n",
	      u8 ( *(res+4+8*i+1) )*65536
	      + u8 (*(res+4+8*i+2) )*256
	      + u8 ( *(res+4+8*i+3) )  );
      printf ( "set blocklen [concat $blocklen {\"%d\"}]\n",
	      u8 ( *(res+4+8*i+5) )*65536
	      + u8 ( *(res+4+8*i+6) )*256
	      + u8 ( *(res+4+8*i+7) ) );
    }
    puts ( "===" );
    fflush ( stdout );
    return;
  }
  
  offset = 4 + 8*block_des;
  
  printf ( "set ps$f %d\n", ( *(res+offset) & 0x0080) != 0 );
  page_len = u8 ( *(res+offset + 1) );

  switch ( *( res + offset ) & 0x003f ) {
  case 0x01:
    switch ( scsi_type ) {
    case TYPE_DISK:
    case TYPE_MOD:
    case TYPE_WORM:
    case TYPE_ROM:
      error_recover ( res+offset, buf, page_len );
      break;
    default:
      break;
    }
    break;

  case 0x02:
    disconnect_param  ( res+offset, buf, page_len );
    break;
    
  case 0x03:
    switch ( scsi_type ) {
    case TYPE_DISK:
    case TYPE_MOD:
      format_param ( res+offset, buf, page_len );
      break;
    case TYPE_PRINTER:
      pprinter_interface_param ( res+offset, buf, page_len );
      break;
    case TYPE_SCANNER:
      measure_unit_param ( res+offset, buf, page_len );
      break;
    default:
      break;
    }
    break;
    
  case 0x04:
    switch ( scsi_type ) {
    case TYPE_DISK:
      drive_param (  res+offset, buf, page_len );
      break;
    case TYPE_PRINTER:
      sprinter_interface_param ( res+offset, buf, page_len );
	break;
    default:
      break;
    }
    break;
    
  case 0x05: 
    switch ( scsi_type ) {
    case TYPE_DISK:
      floppy_param ( res+offset, buf, page_len );
      break;
    case TYPE_PRINTER:
      printer_option_param ( res+offset, buf, page_len );
    default:
      break;
    }
    break;
    
  case 0x06:
    switch ( scsi_type ) {
    case TYPE_WORM:
    case TYPE_MOD:
      optical_param ( res+offset, buf, page_len );
      break;
    default:
      break;
    }
    break;
    
  case 0x07:
    switch ( scsi_type ) {
    case TYPE_DISK:
    case TYPE_WORM:
    case TYPE_MOD:
    case TYPE_ROM:
      verify_error_recovery_param ( res+offset, buf, page_len );
      break;
    default:
      break;
    }
    break;
    
  case 0x08:
    switch ( scsi_type ) {
    case TYPE_DISK:
    case TYPE_WORM:
    case TYPE_MOD:
      cache_control_param ( res+offset, buf, page_len );
      break;
    default:
      break;
    }
    break;
    
  case 0x09:
    peripheral_param ( res+offset, buf, page_len );
    break;
    
  case 0x0a:
    controlmode_param ( res+offset, buf, page_len );
    break;
    
  case 0x0b:
    switch ( scsi_type ) {
    case TYPE_DISK:
    case TYPE_WORM:
    case TYPE_MOD:
    case TYPE_ROM:
      mediatype_param ( res+offset, buf, page_len ); break;
    default:
      break;
    }
    break;

  case 0x0d:
    switch ( scsi_type ) {
    case TYPE_ROM:
      cdrom_param ( res+offset, buf, page_len ); break;
    default:
      break;
    }
    break;

  case 0x0e:
    switch ( scsi_type ) {
    case TYPE_ROM:
      cdrom_audio_param ( res+offset, buf, page_len ); break;
    default:
      break;
    }
    break;
    
  default:
    break;
  }
  puts ( "===" );
  fflush ( stdout );
}



static int media_direct ( int media_type, int device_uniq )
{
  if ( media_type < 0x0080 ) {
    switch ( media_type ) {
    case 0x00:    case 0x01:    case 0x02:    case 0x05:    case 0x06:
    case 0x09:    case 0x0a:    case 0x0d:    case 0x12:    case 0x16:
    case 0x1a:    case 0x1e:    case 0x40:    case 0x44:
      printf ( "set mtype %d\n", media_type ); break;
    default:
      puts ( "set mtype reserved" ); break;
    }
  } else {
    puts ( "set mtype vender" );
  }

  printf ( "set wp %d\n", ( device_uniq & 0x0080 ) != 0 );
  printf ( "set dpofua %d\n",( device_uniq & 0x0010 ) != 0 );
}


/*
   untested
*/
static int media_sequential ( int media_type, int device_uniq )
{
  printf ( "set wp %d\n", ( device_uniq & 0x0080 ) != 0 );

  switch ( (device_uniq & 0x0070) >> 4) {
  case 0x00: puts ( "set bmode output" ); break;
  case 0x01: puts ( "set bmode transfer" ); break;
  case 0x02: puts ( "set bmode otherinit" ); break;
  default:
    puts ( "set bmode reserved" ); break;
  }

  switch ( device_uniq & 0x000f ) {
  case 0x00: puts ( "set speed default" ); break;
  case 0x01: puts ( "set speed slowest"); break;
  default: puts ( "set speed fastest" ); break;
  }
}


/*
   untested
*/
static int density_sequential ( int block, int density )
{
  if ( 0x80 <= density ) {
    puts ( "set density vender" );
  } else {
    switch ( density ) {
    case 0x00:    case 0x04:    case 0x7f:    case 0x01:    case 0x02:
    case 0x03:    case 0x05:    case 0x06:    case 0x07:    case 0x08:
    case 0x09:    case 0x0a:    case 0x0b:    case 0x0c:    case 0x0d:
    case 0x0f:    case 0x10:    case 0x11:    case 0x12:    case 0x13:
    case 0x14:    case 0x15:    case 0x16:    case 0x17:
      printf ( "set density %d\n", density ); break;
    default:
      puts ( "set density reserved" ); break;
    }
  }
}



/*
   untested
*/
static int media_printer ( int media_type, int device_uniq )
{
  switch ( (device_uniq & 0x07) >> 4 ) {
  case 0x00: puts ( "set bmode output" ); break;
  case 0x01: puts ( "set bmode transfer" ); break;
  default:
    puts ( "set buffer reserved" ); break;
  }
}


static int media_optical ( int media_type, int device_uniq )
{
  if ( media_type >= 0x80 ) {
    puts ( "set mtype vender" );
  } else {
    switch ( media_type ) {
    case 0x00:    case 0x01:    case 0x02:    case 0x03:    case 0x04:
    case 0x05:    case 0x06:
      printf ( "set mtype %d\n", media_type ); break;
    default:   puts ( "set mtype reserved" ); break;
    }
  }

  printf ( "set wp %d\n",( device_uniq & 0x0080 ) != 0 );
  printf ( "set dpofua %d\n", ( device_uniq & 0x0010 ) != 0 );
}

static int density_optical ( int i, int density )
{
  if ( density >= 0x0080 ) {
    puts ( "set density vender" ); return;
  }
  switch ( density ) {
  case 0x00:  case 0x01:  case 0x02:  case 0x03:  case 0x04:  case 0x05:
  case 0x06:  case 0x07:  case 0x08:  case 0x09:
    printf ( "set density %d\n", density ); break;
  default:   puts ( "set density reserved" ); break;
  }
}



static int media_cdrom ( int media_type, int device_uniq )
{
  if ( media_type >= 0x0080 ) {
    puts ( "set mtype vender" );
  } else {
    switch ( media_type ) {
    case 0x00:    case 0x01:    case 0x02:    case 0x03:    case 0x05:
    case 0x06:    case 0x07:
      printf ( "set mtype %d\n", media_type ); break;
    default:   puts ( "set mtype reserved" ); break;
    }
  }
  
  printf ( "set dpofua %d\n", ( device_uniq & 0x0010 ) != 0 );
}


static int density_cdrom ( int i, int density )
{
  if ( density >= 0x0080 ) {
    puts ( "set density vender" );
  } else {
    switch ( density ) {
    case 0x00:    case 0x01:    case 0x02:    case 0x03:    case 0x04:
      printf ( "set density %d\n", density ); break;
    default:   puts ( "set density reserved" ); break;
    }
  }
}



  
static int error_recover ( char *page, char *buf, int page_len )
{
  if ( scsi_type != TYPE_ROM ) {
    printf ( "set awre$f %d\n", ( *(page+2) & 0x0080 ) != 0 );
    printf ( "set arre$f %d\n", ( *(page+2) & 0x0040 ) != 0 );
  }

  printf ( "set tb$f %d\n", ( *(page+2) & 0x0020 ) != 0 );
  printf ( "set rc$f %d\n", ( *(page+2) & 0x0010 ) != 0 );

  if ( scsi_type != TYPE_ROM ) {
    printf ( "set eer$f %d\n", ( *(page+2) & 0x0008 ) != 0 );
  }

  printf ( "set per$f %d\n", ( *(page+2) & 0x0004 ) != 0 );
  printf ( "set dte$f %d\n", ( *(page+2) & 0x0002 ) != 0 );
  printf ( "set dcr$f %d\n", ( *(page+2) & 0x0001 ) != 0 );

  printf ( "set rrtc$f %d\n", u8 ( *(page+3) ) );

  if ( scsi_type == TYPE_ROM )  return;

  printf ( "set ecclen$f %d\n", u8 ( *(page+4) ) );
  printf ( "set hofset$f %d\n", (signed int)*(page+5) );
  printf ( "set stbofset$f %d\n", (signed int)*(page+6) );

  if ( page_len <= 6 ) {
    printf ( "set rtlimit$f %d\n", u8 ( *(page+7) ) );
    return;
  }

  printf ( "set wrcount$f %d\n", u8 ( *(page+8) ) );

  if ( scsi_type == TYPE_DISK || scsi_type == TYPE_MOD ) {
    printf ( "set rtlimit$f %d\n", u8 ( *(page+10) )*256 + u8 ( *(page+11) ) );
  }
}



static int disconnect_param ( char *page, char *buf, int page_len )
{
  printf ( "set bfratio$f %f\n", u8( *(page+2) )/255.0*100 );
  printf ( "set beratio$f %f\n", u8( *(page+3) )/255.0*100 );
  printf ( "set bilimit$f %f\n",
	  ( u8 ( *(page+4) )*256 + u8 ( *(page+5) ) )*0.1 );
  printf ( "set dtlimit$f %f\n",
	  ( u8 ( *(page+6) )*256 + u8 ( *(page+7) ) )*0.1 );
  printf ( "set ctlimit$f %f\n",
	  ( u8 ( *(page+8) )*256 + u8 ( *(page+9) ) )*0.1 );

  if ( page_len <= 10 ) return;

  printf ( "set mbsize$f %f\n",
	  (( u8 ( *(page+10) )*256 + u8 ( *(page+11) ) )*512)/1024 );

  printf ( "set dtdc$f %d\n", *(page+12)&0x03 );
}


static int format_param ( char *page, char *buf, int page_len )
{
  printf ( "set tkzone$f %d\n", u8 ( *(page+2) )*256 + u8 ( *(page+3) ) );
  printf ( "set altseczone$f %d\n", u8 ( *(page+4) )*256 + u8 ( *(page+5) ) );
  printf ( "set alttrkzone$f %d\n", u8 ( *(page+6) )*256 + u8 ( *(page+7) ) );
  printf ( "set alttrkunit$f %d\n", u8 ( *(page+8) )*256 + u8 ( *(page+9) ) );
  printf ( "set sectrk$f %d\n", u8 ( *(page+10) )*256 + u8 ( *(page+11) ) );
  printf ( "set bytesec$f %d\n", u8 ( *(page+12) )*256 + u8 ( *(page+13) ) );
  printf ( "set interleave$f %d\n",
	  u8 ( *(page+14) )*256 + u8 ( *(page+15) ) );
  printf ( "set trkskew$f %d\n", u8 ( *(page+16) )*256 + u8 ( *(page+17) ) );
  printf ( "set cylskew$f %d\n", u8 ( *(page+18) )*256 + u8 ( *(page+19) ) );
  printf ( "set shsec$f %d\n", u8 (*(page+20)) >> 6 );
  printf ( "set rmb$f %d\n", ( *(page+20) & 0x0020 ) != 0 );
  printf ( "set surf$f %d\n", ( *(page+20) & 0x0010 ) != 0 );
}



/*
   untested
*/
static int pprinter_interface_param  ( char *page, char *buf, int page_len )
{
  return;
  puts ( "wm title .msub {Parallel Printer Interface parameter}" );

  puts ( "label .msub.frame0.l.ps -text {ѥƥ쥯} -relief groove\n"
	"pack .msub.frame0.l.ps -fill x" );
  fputs  ( "label .msub.frame0.r.ps -relief groove -text ", stdout );
  switch ( (*(page+2) & 0x00c0) >> 6 ) {
  case 0x00: puts ( "{Non Parity}" ); break;
  case 0x01: puts ( "{Even Parity}" ); break;
  case 0x02: puts ( "{Odd Parity}" ); break;
  case 0x03: puts ( "{(reserved)}" ); break;
  }
  puts ( "pack .msub.frame0.r.ps -fill x" );

  puts ( "label .msub.frame0.l.pipc -text {Paper Instruction ѥƥ˴ޤ} -relief groove\n"
	"pack .msub.frame0.l.pipc -fill x" );
  printf ( "label .msub.frame0.r.pipc -text %s -relief groove\n",
	  yesno ( *(page+2) & 0x20 ) );
  puts ( "pack .msub.frame0.r.pipc -fill x" );

  puts ( "label .msub.frame0.l.vcbp -text {VFU ȥХȤǤΥȥӥåȤζ} -relief groove\n"
	"pack .msub.frame0.l.vcbp -fill x" );
  printf ( "label .msub.frame0.r.vcbp -text {%s} -relief groove\n",
	  (*(page+2) & 0x08) ? "VFU channel True, LF False" :
	  "VFU channel False, LF True" );
  puts ( "pack .msub.frame0.r.vcbp -fill x" );

  puts ( "label .msub.frame0.l.vcbs -text {VFU ȥХȤǤΥȥӥåȤΰ} -relief groove\n"
	"pack .msub.frame0.l.vcbs -fill x" );
  printf ( "label .msub.frame0.r.vcbs -text {%s} -relief groove\n",
	  ( *(page+2) & 0x04 ) ? 
	  "C = bit 6, EE = bit 5 - 4" : "C = bit 4, EE = bit 6 - 5" );
  puts ( "pack .msub.frame0.r.vcbs -fill x" );

  puts ( "label .msub.frame0.l.ves -text {饤󥹥å׻ؼ EE ӥå} -relief groove\n"
	"pack .msub.frame0.l.ves -fill x" );
  printf ( "label .msub.frame0.r.ves -text %s -relief groove\n",
	  yesno ( *(page+2) & 0x02 ) );
  puts ( "pack .msub.frame0.r.ves -fill x" );

  puts ( "label .msub.frame0.l.autofd -text {Auto Line Feed ΰ} -relief groove\n"
	"pack .msub.frame0.l.autofd -fill x" );
  printf ( "label .msub.frame0.r.autofd -text {%s} -relief groove\n",
	  ( *(page+2) & 0x01 ) ? "Assert" : "Negate" );
  puts ( "pack .msub.frame0.r.autofd -fill x" );
}



/*
   untested
*/
static int measure_unit_param  ( char *page, char *buf, int page_len )
{
  char unit[64];

  return;
  puts ( "wm title .msub {Measurement Unit Parameter}" );

  switch ( *(page+2) ) {
  case 0x00: strcpy ( unit, "inch" ); break;
  case 0x01: strcpy ( unit, "millimeter" ); break;
  case 0x02: strcpy ( unit, "point" ); break;
  default:
    strcpy ( unit, "(reserved)" ); break;
  }

  puts ( "label .msub.frame0.l.bmu -text {ñ} -relief groove\n"
	"pack .msub.frame0.l.bmu -fill x" );
  printf ( "label .msub.frame0.r.bmu -text {%s} -relief groove\n", unit );
  puts ( "pack .msub.frame0.r.bmu -fill x" );

  puts ( "label .msub.frame0.l.res -text {} -relief groove\n"
	"pack .msub.frame0.l.res -fill x" );
  printf ( "label .msub.frame0.r.res -text {%d dot/%s} -relief groove\n",
	  u8(*(page+4))*256 + u8(*(page+5)), unit );
  puts ( "pack .msub.frame0.r.res -fill x" );
}



static int drive_param  ( char *page, char *buf, int page_len )
{
  int lz;

  printf ( "set cyl$f %d\n",
	  u8 ( *(page+2) )*65536
	  + u8 ( *(page+3) )*256 + u8 ( *(page+4) ) );
  printf ( "set head$f %d\n", u8 ( *(page+5) ) );
  printf ( "set wpadj$f %d\n",
	  u8 ( *(page+6) )*65536
	  + u8 ( *(page+7) )*256 + u8 ( *(page+8) ) );
  printf ( "set wcrdc$f %d\n",
	  u8 ( *(page+9) )*65536
	  + u8 ( *(page+10) )*256 + u8 ( *(page+11) ) );
  printf ( "set strate$f %d\n",
	  ( u8 ( *(page+12) )*256 + u8 ( *(page+13) ) )*10 );

  lz = u8 ( *(page+14) )*65536 + u8 ( *(page+15) )*256 + u8 ( *(page+16) );
  if ( lz > 0x007fffff ) lz = 0x01000000 - lz;

  printf ( "set lzone$f %d\n", lz );

  if ( page_len <= 18 )  return;

  printf ( "set rpl$f %d\n", ( *(page+17) & 0x03 ) );
  printf ( "set rofset$f %f\n", u8 ( *(page+19) )/256 );
  printf ( "set rpm$f %d\n", u8 ( *(page+20) )*256 + u8 ( *(page+21) ) );
}



/*
   untested
*/
static int sprinter_interface_param  ( char *page, char *buf, int page_len )
{
  return;
  puts ( "wm title .msub {Serial Printer Interface Parameter}" );

  puts ( "label .msub.frame0.l.sp -text {ȥåץӥåĹ} -relief groove\n"
	"pack .msub.frame0.l.sp -fill x" );
  printf ( "label .msub.frame0.r.sp -text {%d} -relief groove\n",
	  *(page + 2) & 0x003f );
  puts ( "pack .msub.frame0.r.sp -fill x" );

  
  puts ( "label .msub.frame0.l.parity -text {ѥƥ} -relief groove\n"
	"pack .msub.frame0.l.parity -fill x" );
  fputs  ( "label .msub.frame0.r.parity -relief groove -text ", stdout );
  switch ( (*(page+3) & 0x00e0) >> 5 ) {
  case 0x00: puts ( "{Non Parity}" ); break;
  case 0x01: puts ( "{Mark}" ); break;
  case 0x02: puts ( "{Space}" ); break;
  case 0x03: puts ( "{Even Parity}" ); break;
  case 0x04: puts ( "{Odd Parity}" ); break;
  default:
    puts ( "{(reserved)}" ); break;
  }
  puts ( "pack .msub.frame0.r.parity -fill x" );

  puts ( "label .msub.frame0.l.bpc -text {ӥåȡ饯} -relief groove\n"
	"pack .msub.frame0.l.bpc -fill x" );
  printf ( "label .msub.frame0.r.bpc -text {%d} -relief groove\n",
	  *(page + 3) & 0x000f );
  puts ( "pack .msub.frame0.r.bpc -fill x" );

  puts ( "label .msub.frame0.l.rts -text {RTS ΰ} -relief groove\n"
	"pack .msub.frame0.l.rts -fill x" );
  printf ( "label .msub.frame0.r.rts -text {%s} -relief groove\n",
	  ( *(page + 4) & 0x0080 ) ? "DTR ˽" : " High" );
  puts ( "pack .msub.frame0.l.rts -fill x" );

  puts ( "label .msub.frame0.l.cts -text {CTS ͭ} -relief groove\n"
	"pack .msub.frame0.l.rts -fill x" );
  printf ( "label .msub.frame0.r.cts -text %s -relief groove\n",
	  yesno ( *(page + 4) & 0x0040 ) );
  puts ( "pack .msub.frame0.r.rts -fill x" );

  puts ( "label .msub.frame0.l.pacing -text {ϥɥ} -relief groove\n"
	"pack .msub.frame0.l.pacing -fill x" );
  fputs  ( "label .msub.frame0.r.pacing -relief groove -text ", stdout );
  switch ( *(page+4) & 0x000f ) {
  case 0x00: puts ( "{No Protocol}" ); break;
  case 0x01: puts ( "{XON/XOFF}" ); break;
  case 0x02: puts ( "{ETX/ACK}" ); break;
  case 0x03: puts ( "{DTR}" ); break;
  case 0x04: case 0x05: case 0x06: case 0x07:
    puts ( "{(reserved)}" ); break;
  default:
    puts ( "{Vender Unique}" ); break;
  }
  puts ( "pack .msub.frame0.r.pacing -fill x" );

  puts ( "label .msub.frame0.l.bps -text {ӥåȥ졼} -relief groove\n"
	"pack .msub.frame0.l.bps -fill x" );
  printf ( "label .msub.frame0.r.bps -text {%d bit/sec} -relief groove\n",
	  u8 ( *(page+5) )*65536
	  + u8 ( *(page+6) )*256
	  + u8 ( *(page+7) ) );
  puts ( "pack .msub.frame0.r.bps -fill x" );

}



/*
   untested
*/
static int floppy_param ( char *page, char *buf, int page_len )
{
  return;
  
  puts ( "wm title .msub {Floppy Disk Parameter}" );
  
  printf ( "--- Data Transfer Rate:                 %d kbit/sec\n",
	  u8 ( *(page+2) )*256
	  +	u8 ( *(page+3) ) );
  printf ( "--- Heads:                                  %d\n",
	  u8 ( *(page+4) ) );
  printf ( "--- Sectors/Track:                          %d\n",
	  u8 ( *(page+5) ) );
  printf ( "--- Bytes/Sector:                           %d byte",
	  u8 ( *(page+6) )*256
	  +	u8 ( *(page+7) ) );
  printf ( "--- Cylinders:                              %d\n",
	  u8 ( *(page+8) )*256
	  +	u8 ( *(page+9) ) );
  printf ( "--- Write Phase Adjustment Start Cylinder:  %d\n",
	  u8 ( *(page+10) )*256
	  +	u8 ( *(page+11) ) );
  printf ( "--- Write Current Reduction Start Cylinder: %d\n",
	  u8 ( *(page+12) )*256
	  +	u8 ( *(page+13) ) );
  printf ( "--- Drive Step Rate:                        %f msec\n",
	  ( u8 ( *(page+14) )*256
	   +	u8 ( *(page+15) ) )*0.1 );
  printf ( "--- Drive Step Pulse Width:                 %d usec\n",
	  u8 ( *(page+16) ) );
  printf ( "--- Head settling Delay:                    %f msec\n",
	  ( u8 ( *(page+17) )*256
	   +	u8 ( *(page+18) ) )*0.1 );
  printf ( "--- Motor On Delay:                         %f sec\n",
	  u8 ( *(page+19) )*0.1 );
  printf ( "--- Motor Off Delay:                        %f sec\n",
	  u8 ( *(page+20) )*0.1 );
  printf ( "--- Exist `Media Access Ready' Line:        %s\n",
	  yesno ( *(page+21) & 0x0080 ) );
  printf ( "--- Sector Number Count From:               %d \n",
	  ( *(page+21) & 0x0040 ) ? 1 : 0  );
  printf ( "--- Motor On Line:                          %s\n",
	  (*(page+21) & 0x0020) ? "Keep Release" : "Assert" );
  printf ( "--- Step Pulse Addition / 1 Cylinder Seek:  %d\n",
	  *(page+22) & 0x000f );
  printf ( "--- Write Phase Adjust:                     %d\n",
	  u8 ( *(page+23) ) );
  if ( page_len <= 21 ) return;
  printf ( "--- Head Load Delay:                        %d msec\n",
	  u8 ( *(page+24) ) );
  printf ( "--- Head Unload Delay:                      %d msec\n",
	  u8 ( *(page+25) ) );
  printf ( "--- pin 34 means:              %s, ",
	  ( *(page+26) & 0x0080 ) ?
	  "active high" : "active low " );
  switch ( ( *(page+26) & 0x0070 )>> 4 ) {
  case 0x00: puts ( "open" ); break;
  case 0x01: puts ( "Ready" ); break;
  case 0x02: puts ( "Disk Changed" ); break;
  default:
    puts ( "(undefined)" ); break;
  }
  printf ( "--- pin 2 :                                 $%02X (undefined)\n",
	  *(page+26) & 0x000f );
  printf ( "--- pin 4:                     %s, ",
	  ( *(page+27) & 0x0080 ) ?
	  "active high" : "active low" );
  switch ( ( *(page+27) & 0x0070 )>> 4 ) {
  case 0x00: puts ( "Open" ); break;
  case 0x01: puts ( "In Use" ); break;
  case 0x02: puts ( "Eject" ); break;
  case 0x03: puts ( "Head Load" ); break;
  default:
    puts ( "(undefined)" ); break;
  }
  printf ( "--- pin 1:                     %s, ",
	  ( *(page+27) & 0x0008 ) ?
	  "active high" : "active low" );
  switch (  *(page+27) & 0x0070 ) {
  case 0x00: puts ( "open" ); break;
  case 0x01: puts ( "Disk Change Reset" ); break;
  default:
    puts ( "(undefined)" ); break;
  }
  printf ( "--- Media Rotational Speed:                 %d rpm\n",
	  u8 ( *(page+28) )*256
	  + u8 ( *(page+29) )  );
}  
  


/*
   untested
*/
static int printer_option_param ( char *page, char *buf, int page_len )
{
  int fontid;

  return;
  puts ( "=== Printer Option Parameter" );
  puts ( buf );
  printf ( "--- Electronic VFU:              %s\n",
	  yesno (*(page+2) & 0x0080 ) );
  fontid = *(page+2) & 0x007f;
  fputs  ( "--- font ID:                     ", stdout );
  if ( fontid == 0x00 ) 
    puts ( "Default Font" );
  else if ( 0x01 <= fontid && fontid <= 0x003f ) 
    printf ( "$%02X (reserved)\n", fontid );
  else
    printf ( "$%02X Vender Unique\n", fontid );

  fputs ( "--- Slew Mode:                    ", stdout );
  switch ( ( *(page+3)& 0x0030 ) >> 4 ) {
  case 0x00: puts ( "Slew and Print" ); break;
  case 0x01: puts ( "VFU Init, then Slew and Print" ); break;
  case 0x10: puts ( "Slew and Print ( EVFU )" ); break;
  case 0x11: puts ( "(reserved)" ); break;
  default:
    puts ( "***** warning! some bugs in this program..." );
    break;
  }
  printf ( "--- Step Count Truncate Enable:  %s\n",
	  yesno ( *(page+3) & 0x02 ) );
  printf ( "--- ASCII Forms Control:         %s\n",
	  yesno ( *(page+3) & 0x01 ) );
  printf ( "--- Maximum Line Length:         %d\n",
	  u8 ( *(page+4) )*256 + u8 ( *(page+5) ) );
  printf ( "--- EVFU Format Start Character: $%02X\n",
	  u8 ( *(page+6) ) );
  printf ( "--- EVFU Format Stop Character:  $%02X\n",
	  u8 ( *(page+7) ) );
  fputs  ( "--- Line Slwe Option:            ", stdout );
  switch ( ( *(page+8) & 0x00f0 ) >> 4 ) {
  case 0x00: puts ( "unsupported" ); break;
  case 0x01: puts ( "CR" ); break;
  case 0x02: puts ( "LF" ); break;
  case 0x03: puts ( "CR/LF" ); break;
  case 0x04: case 0x05: case 0x06: case 0x07:
    puts ( "(reserved)" ); break;
  default:
    puts ( "Vender Unique" ); break;
  }
  fputs  ( "--- Form Slwe Option:            ", stdout );
  switch ( *(page+8) & 0x000f ) {
  case 0x00: puts ( "unsupported" ); break;
  case 0x01: puts ( "FF" ); break;
  case 0x02: puts ( "CR/FF" ); break;
  case 0x03: case 0x04: case 0x05: case 0x06: case 0x07:
    puts ( "(reserved)" ); break;
  default:
    puts ( "Vender Unique" ); break;
  }
  fputs ( "--- Data Termination Option:      ", stdout );
  switch ( (*(page+9) & 0x00f0)>> 4 ) {
  case 0x00: puts ( "Default" ); break;
  case 0x01: puts ( "Do Nothing" ); break;
  case 0x02: puts ( "CR" ); break;
  case 0x03: puts ( "LF" ); break;
  case 0x04: puts ( "CR/LF" ); break;
  case 0x05: puts ( "FF" ); break;
  case 0x06: puts ( "CR/FF" ); break;
  case 0x07: puts ( "`Zero Line Slew' Command" ); break;
  case 0x08:  case 0x09:   case 0x0a:   case 0x0b:
    puts ( "(reserved)" ); break;
  default:
    puts ( "Vender Unique" ); break;
  }
}



static int optical_param ( char *page, char *buf, int page_len )
{
  printf ( "set uread %d\n", ( *(page+2) & 0x01 ) != 0 );
}



static int verify_error_recovery_param ( char *page, char *buf, int page_len )
{
  printf ( "set eer$f %d\n", ( *(page+2) & 0x0008 ) != 0 );
  printf ( "set per$f %d\n", ( *(page+2) & 0x0004 ) != 0 );
  printf ( "set dte$f %d\n", ( *(page+2) & 0x0002 ) != 0 );
  printf ( "set dcr$f %d\n", ( *(page+2) & 0x0001 ) != 0 );
  printf ( "set vrtc$f %d\n",  u8 ( *(page+3) ) );
  printf ( "set ecclen$f %d\n", u8 ( *(page+4) ) );
  printf ( "set rtlimit$f %d\n", u8 ( *(page+10) )*256 + u8 ( *(page+11) ) );
}




static int cache_control_param ( char *page, char *buf, int page_len )
{
  printf ( "set wce$f %d\n", ( *(page+2) & 0x04 ) != 0 );
  printf ( "set mf$f %d\n", ( *(page+2) & 0x02 ) != 0 );
  printf ( "set rcd$f %d\n", ( ( *(page+2) & 0x01 ) ) != 0 );
  
  printf ( "set drrp$f %d\n", ( *(page+3) & 0x00f0 ) >> 4 );
  printf ( "set wrp$f %d\n",  *(page+3) & 0x0f );
  printf ( "set dptl$f %d\n", u8 ( *(page+4) )*256 + u8 ( *(page+5) ) );
  printf ( "set minp$f %d\n", u8 ( *(page+6) )*256 + u8 ( *(page+7) ) );
  printf ( "set maxp$f %d\n", u8 ( *(page+8) )*256 + u8 ( *(page+9) ) );
  printf ( "set maxc$f %d\n", u8 ( *(page+10) )*256 + u8 ( *(page+11) ) );
}

  
/*
   untested
*/
static int peripheral_param ( char *page, char *buf, int page_len )
{
  int id;

  return;

  puts ( "=== Peripheral Device Parameter" );
  puts ( buf );
  fputs ( "--- Interface ID: ", stdout );
  id = u8 ( *(page+2) )*256 + u8 ( *(page+3) );
  if ( id >= 0x008000 ) {
    puts ( "vender unique" );
    return;
  }
  switch ( id ) {
  case 0x0000: puts ( "SCSI" ); break;
  case 0x0001: puts ( "SMD" ); break;
  case 0x0002: puts ( "E-SMD" ); break;
  case 0x0003: puts ( "IPI-2" ); break;
  case 0x0004: puts ( "IPI-3" ); break;
  default:     puts ( "(reserved)" ); break;
  }
}


static int controlmode_param ( char *page, char *buf, int page_len )
{
  printf ( "set rlec$f %d\n", ( *(page+2) & 0x01 ) != 0 );
  printf ( "set qam$f %d\n", (*(page+3)&0x00f0)>>4 );
  printf ( "set qerr$f %d\n", ( *(page+3) & 0x02 ) != 0 );
  printf ( "set dque$f %d\n", ( *(page+3) & 0x01 ) != 0 );
  printf ( "set eeca$f %d\n", ( *(page+4) & 0x0080 ) != 0 );
  printf ( "set raenp$f %d\n",( *(page+4) & 0x04 ) != 0 );
  printf ( "set uaaenp$f %d\n", ( *(page+4) & 0x02 ) != 0 );
  printf ( "set eaenp$f %d\n", ( *(page+4) & 0x01 ) != 0 );
  printf ( "set raent$f %d\n", u8 (*(page+6))*256 + u8 (*(page+7)) );
}


static int mediatype_param ( char *page, char *buf, int page_len )
{
  printf ( "set media1$f %d\n", u8 ( *(page+4) ) );
  printf ( "set media2$f %d\n", u8 ( *(page+5) ) );
  printf ( "set media3$f %d\n", u8 ( *(page+6) ) );
  printf ( "set media4$f %d\n", u8 ( *(page+7) ) );
}

static int cdrom_param ( char *page, char *buf, int page_len )
{
  printf ( "set inactivity$f %d\n", *(page+3) & 0x0f );
  printf ( "set msfsm$f %d\n", u8(*(page+4))*256 + u8(*(page+5)) );
  printf ( "set msffs$f %d\n", u8(*(page+6))*256 + u8(*(page+7)) );
}
  

static int cdrom_audio_param ( char *page, char *buf, int page_len )
{
  int i;

  printf ( "set immed$f %d\n", ( *(page+2) & 0x04 ) != 0 );
  printf ( "set sotc$f %d\n", ( *(page+2) & 0x02 ) != 0 );
  printf ( "set aprval$f %d\n", ( *(page+5) & 0x0080 ) != 0 );
  printf ( "set blocksec$f %d\n", *(page+5) & 0x000f );
  printf ( "set bsunit$f %d\n", *(page+5) & 0x000f );
  printf ( "set blocksec$f %d\n", u8(*(page+6))*256 + u8(*(page+7)) );

  for ( i = 0; i < 4; ++i ) {
    printf ( "set mcdaudio_psel${f}(0_%d) %d\n",
	    i, (*(page+8) & (1<<i)) != 0 );
    printf ( "set mcdaudio_psel${f}(1_%d) %d\n",
	    i, (*(page+10) & (1<<i)) != 0 );
    printf ( "set mcdaudio_psel${f}(2_%d) %d\n",
	    i, (*(page+12) & (1<<i)) != 0 );
    printf ( "set mcdaudio_psel${f}(3_%d) %d\n",
	    i, (*(page+14) & (1<<i)) != 0 );
  }

  printf ( "set pvol${f}(0) %d\n", u8(*(page+9)) );
  printf ( "set pvol${f}(1) %d\n", u8(*(page+11)) );
  printf ( "set pvol${f}(2) %d\n", u8(*(page+13)) );
  printf ( "set pvol${f}(3) %d\n", u8(*(page+15)) );
}
