/*
   readdefect.c: send Read Defect Data command
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/../scsi/scsi.h>
#include "var.h"

#define P_LIST            0x10
#define G_LIST            0x08

#define DEFECT_BLOCK      0     /* block address (unrecommended) */
#define DEFECT_RELATIVE   4     /* index byte distance */
#define DEFECT_PHYSICAL   5     /* physical sector address */
#define DEFECT_VENDER     6     /* vender unique */

static int showlist ( int, char * );

void read_defect ( int pg, int fmt, char *fname )
{
  int status, fd;
  struct sg_header *sg;
  char *cdb = bufcdb + SCMD_HEADSIZ;
  char *res = bufres + SCMD_HEADSIZ;

  fd = open ( fname, O_RDWR );
  if ( fd < 0 ) {
    puts ( "*** 001 CAN'T OPEN" );
    fflush ( stdout );
    close ( fd );
    return;
  }

  *(cdb +  0) = READ_DEFECT_DATA;
  *(cdb +  1) = scsi_lun << 5;
  if ( pg == 'p' )
    *(cdb +  2) = P_LIST;
  else
    *(cdb +  2) = G_LIST;
  switch ( fmt ) {
  case 'b': /* ֥åɥ쥹 */
    *(cdb + 2) |= DEFECT_BLOCK;
    break;
  case 'r': /* ǥåΥХȵΥ */
    *(cdb + 2) |= DEFECT_RELATIVE;
    break;
  case 'p': /* ʪɥ쥹 */
    *(cdb + 2) |= DEFECT_PHYSICAL;
    break;
  default:
    break;
  }
  *(cdb +  3) = 0;
  *(cdb +  4) = 0;
  *(cdb +  5) = 0;
  *(cdb +  6) = 0;
  *(cdb +  7) = SCMD_RECVMAX >> 8;
  *(cdb +  8) = SCMD_RECVMAX & 0x00ff;
  *(cdb +  9) = 0;

  status = scsicmd ( fd, bufres, SCMD_RECVMAX, bufcdb, SCMD_HEADSIZ + 10 );
  close ( fd );

  if ( status ) {
    sg = (struct sg_header *)bufres;
    puts ( "*** 002 FAIL TO SEND COMMAND" );
    printf ( "set status %d\n", status );
    printf ( "set errcode %d\n", sg->sense_buffer[0] & 0x007f );
    printf ( "set sensekey %d\n", sg->sense_buffer[2] & 0x000f );
    printf ( "set asc %d\n", u8 (sg->sense_buffer[12]) );
    printf ( "set ascq %d\n", u8 (sg->sense_buffer[13]) );
    puts ( "===" );
    fflush ( stdout );
    return;
  }
  
  showlist ( pg, res );
}
  


static int showlist ( int pg, char *res )
{
  int form    = u8 ( *(res+1) & 0x07 );
  int tablesiz = u8 ( *(res+2) )*256 + u8 ( *(res+3) );
  int listlen = tablesiz + 3;
  int siz, i;

  switch ( form ) {
  case DEFECT_BLOCK:
    puts ( "set form b" );
    siz = 4; break;
  case DEFECT_RELATIVE:
    puts ( "set form r" );
    siz = 8; break;
  case DEFECT_PHYSICAL:
    puts ( "set form p" );
    siz = 8; break;
  default:
    puts ( "*** 003 UNSUPPORT FORM" );
    return -1;
  }

  printf ( "set %ccount %d\n", pg, tablesiz/siz );
  /* printf ( ".defect.framepg.frame%c.list%c delete 0 end\n", pg, pg ); */
  for ( i = 4; i < listlen; i += siz ) {
    switch ( form ) {
    case DEFECT_BLOCK:
      printf ( "set %clist [concat $%clist {\"%ld\"}]\n",
	      pg, pg,
	      65536L*256*u8( *(res+i) ) + 65536L*u8( *(res+i+1) )
	      + 256L*u8( *(res+i+2) ) + 1L*u8( *(res+i+3) ) );
      break;
    case DEFECT_RELATIVE:
      printf ( "set %clist [concat $%clist {\"%8d/%3d/%10ld\"}]\n",
	      pg, pg,
	      65536*u8( *(res+i) ) + 256*u8( *(res+i+1) ) + u8( *(res+i+2) ),
	      u8( *(res+i+3) ),
	      65536L*256*u8( *(res+i+4) ) + 65536L*u8( *(res+i+5) )
	      + 256L*u8( *(res+i+6) ) + 1L*u8( *(res+i+7) ) );
      break;
    case DEFECT_PHYSICAL:
      printf ( "set %clist [concat $%clist {\"%8d/%3d/%10ld\"}]\n",
	      pg, pg,
	      65536*u8( *(res+i) ) + 256*u8( *(res+i+1) ) + u8( *(res+i+2) ),
	      u8( *(res+i+3) ),
	      65536L*256*u8( *(res+i+4) ) + 65536L*u8( *(res+i+5) )
	      + 256L*u8( *(res+i+6) ) + 1L*u8( *(res+i+7) ) );
      break;
    default:
      return -1;
    }
  }
  puts ( "===" );
  fflush ( stdout );
}
