/*
   rwbuffer.c: read/write buffer benchmark
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/../scsi/scsi.h>
#include <unistd.h>
#include <signal.h>
#include "var.h"
#include "rwbuffer.h"
#include "scsicmd.h"

#define BUFLEN ( 4096 - SCMD_HEADSIZ - 10 )
static int sigcatch;

static long x_rwbuffer_info ( int );
static void sighandle ( int );
static void readbench ( int , int *, int *, int *, int *, int *, int *,
		       int, int );
static void writebench ( int , int *, int *, int *, int *, int *, int *,
			int, int );

/*
   rwbuffer_info: get buffer infomation
*/
void rwbuffer_info ( char *fname )
{
  int status, addlen, fd;
  struct sg_header *sg = (struct sg_header *)bufres;
  char *cdb = bufcdb + SCMD_HEADSIZ;
  char *res = bufres + SCMD_HEADSIZ;

  fd = open ( fname, O_RDWR );
  if ( fd < 0 ) {
    puts ( "*** 001 CAN'T OPEN" );
    fflush ( stdout );
    return;
  }

  *(cdb +  0) = READ_BUFFER;
  *(cdb +  1) = (scsi_lun << 5) | RWBUFFER_HEAD_DATA;
  *(cdb +  2) = 0;
  *(cdb +  3) = 0;
  *(cdb +  4) = 0;
  *(cdb +  5) = 0;
  *(cdb +  6) = 0;
  *(cdb +  7) = 0;
  *(cdb +  8) = 4;
  *(cdb +  9) = 0;

  status = scsicmd ( fd, bufres, SCMD_RECVMAX, bufcdb, SCMD_HEADSIZ + 10 );
  close ( fd );

  if ( status ) {
    puts ( "*** 002 FAIL TO SEND COMMAND" );
    printf ( "set status %d\n", status );
    printf ( "set errcode %d\n", sg->sense_buffer[0] & 0x007f );
    printf ( "set sensekey %d\n", sg->sense_buffer[2] & 0x000f );
    printf ( "set asc %d\n", u8 (sg->sense_buffer[12]) );
    printf ( "set ascq %d\n", u8 (sg->sense_buffer[13]) );
    puts ( "===" );
    fflush ( stdout );
    return;
  }

  printf ( "set offset_boundary %d\n", u8 ( *(res + 0) ) );
  printf ( "set buffer_capacity %d\n",
	  ( u8 ( *(res + 1 ) ) * 65536 ) |
	  ( u8 ( *(res + 2) ) *256 ) |
	   ( u8 ( *(res + 3) ) ) );
  puts ( "===" );
  fflush ( stdout );
}


/*
   x_rwbuffer_info: internal version of rwbuffer_info
*/
long x_rwbuffer_info ( int fd )
{
  int status, addlen;
  struct sg_header *sg = (struct sg_header *)bufres;
  char *cdb = bufcdb + SCMD_HEADSIZ;
  char *res = bufres + SCMD_HEADSIZ;

  *(cdb +  0) = READ_BUFFER;
  *(cdb +  1) = (scsi_lun << 5) | RWBUFFER_HEAD_DATA;
  *(cdb +  2) = 0;
  *(cdb +  3) = 0;
  *(cdb +  4) = 0;
  *(cdb +  5) = 0;
  *(cdb +  6) = 0;
  *(cdb +  7) = 0;
  *(cdb +  8) = 4;
  *(cdb +  9) = 0;

  status = scsicmd ( fd, bufres, SCMD_RECVMAX, bufcdb, SCMD_HEADSIZ + 10 );

  if ( status ) 
    return -1;

  return ( u8 ( *(res + 1 ) ) * 65536 ) |
	  ( u8 ( *(res + 2) ) *256 ) |
	   ( u8 ( *(res + 3) ) );
}



/*
   rbuffer_bench: Read Buffer/Write Buffer benchmark
*/
void rwbuffer_bench ( char *fname, int mode, int sec )
{
  int status, addlen, fd;
  int buflen, buflen4;
  int count;
  int laststat, errcode, sensekey, asc, ascq;
  struct sg_header *sg = (struct sg_header *)bufres;
  char *cdb = bufcdb + SCMD_HEADSIZ;
  char *res = bufres + SCMD_HEADSIZ;
  
  fd = open ( fname, O_RDWR );
  if ( fd < 0 ) {
    puts ( "*** 001 CAN'T OPEN" );
    fflush ( stdout );
    return;
  }

  buflen = x_rwbuffer_info ( fd );
  if ( buflen < 0 ) { /* bad range */
    close ( fd );
    puts ( "*** 003 NOT SUPPORTED" );
    fflush ( stdout );
    return;
  }

  buflen = ( buflen > BUFLEN ) ? BUFLEN : buflen;
  buflen4 = buflen + 4;

  *(cdb +  0) = ( mode == 'W' ) ?  WRITE_BUFFER : READ_BUFFER;
  *(cdb +  1) = (scsi_lun << 5) | RWBUFFER_HEAD_DATA;
  *(cdb +  2) = 0;
  *(cdb +  3) = 0;
  *(cdb +  4) = 0;
  *(cdb +  5) = 0;
  *(cdb +  6) = u8 ( buflen4 >> 16 );
  *(cdb +  7) = u8 ( buflen4 >> 8 );
  *(cdb +  8) = u8 ( buflen );
  *(cdb +  9) = 0;

  sigcatch = 0;
  laststat = errcode = sensekey = asc = ascq = 0;

  if ( mode == 'W' )
    writebench ( fd, &count, &laststat, &errcode, &sensekey, &asc, &ascq,
		buflen4, sec );
  else
    readbench ( fd, &count, &laststat, &errcode, &sensekey, &asc, &ascq,
	       buflen4, sec );

  close ( fd );

  printf ( "set buflen   %d\n", buflen4 );
  printf ( "set count    %d\n", count );
  printf ( "set status   %d\n", laststat );
  printf ( "set errcode  %d\n", errcode );
  printf ( "set sensekey %d\n", sensekey );
  printf ( "set asc      %d\n", asc );
  printf ( "set ascq     %d\n", ascq );
  puts ( "===" );
  fflush ( stdout );
}


static void sighandle ( int sig )
{
  if ( sig != SIGALRM ) {
    return; /* bad behavior ... */
  }
  sigcatch = 1;
}



void readbench ( int fd, int *count, int *laststat, int *errcode,
		int *sensekey, int *asc, int *ascq, int buflen4, int sec )
{
  int status;
  struct sg_header *sg = (struct sg_header *)bufres;

  signal ( SIGALRM, &sighandle );
  alarm ( sec );
  for ( *count = 0; !sigcatch; ++*count ) {
    status = scsicmd ( fd, bufres, SCMD_HEADSIZ + buflen4,
		      bufcdb, SCMD_HEADSIZ + 10 );
    switch ( status ) {
    case SCMD_NORM_END:
    case SCMD_SEND_ERR:
    case SCMD_RECV_ERR:
      break;
    default:
      *laststat = status;
      *errcode = sg->sense_buffer[0] & 0x007f;
      *sensekey = sg->sense_buffer[2] & 0x000f;
      *asc = u8 (sg->sense_buffer[12]);
      *ascq = u8 (sg->sense_buffer[13]);
      break;
    }
  }
}



void writebench ( int fd, int *count, int *laststat, int *errcode,
		int *sensekey, int *asc, int *ascq, int buflen4, int sec )
{
  int status;
  struct sg_header *sg = (struct sg_header *)bufres;

  signal ( SIGALRM, &sighandle );
  alarm ( sec );
  for ( *count = 0; !sigcatch; ++*count ) {
    status = scsicmd ( fd, bufres, SCMD_HEADSIZ,
		      bufcdb, SCMD_HEADSIZ + buflen4 );
    switch ( status ) {
    case SCMD_NORM_END:
    case SCMD_SEND_ERR:
    case SCMD_RECV_ERR:
      break;
    default:
      *laststat = status;
      *errcode = sg->sense_buffer[0] & 0x007f;
      *sensekey = sg->sense_buffer[2] & 0x000f;
      *asc = u8 (sg->sense_buffer[12]);
      *ascq = u8 (sg->sense_buffer[13]);
      break;
    }
  }
}
