/*
   scsicmd.c: handle SCSI command

   1996 (C) by JR0BAK
*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <linux/../scsi/sg.h>

#include "scsicmd.h"

/*
   scsicmd: send SCSI command

   argument:
         fd:          file handle to SCSI generic file
         recvbuf:     result code buffer pointer
	 recvlen:                        size      
	 sendbuf:     send buffer pointer
	 sendlen:                 size

   function return code: status code
	     0 ... normal end
	 not 0 ... error code ( defined at scsicmd.h )
*/
scsicmd ( int fd, void *recvbuf, int recvlen,
	     void *sendbuf, int sendlen )
{
  struct sg_header *sg;
  int wstatus, rstatus;
  char *emsg;
  int i;

  if ( sendbuf == NULL || sendlen <= SCMD_HEADSIZ
      || recvbuf == NULL || recvlen < SCMD_HEADSIZ
      || recvlen > SCMD_RECVMAX ) {
    return SCMD_PARM_ERR;
  }

  sg = (struct sg_header *)sendbuf;
  sg->pack_len = sendlen;
  sg->reply_len = recvlen;
  sg->result = 0;
  sg->twelve_byte = 0;

  wstatus = write ( fd, sendbuf, sendlen );
  if ( wstatus < sendlen ) {
    return SCMD_SEND_ERR;
  }

  rstatus = read ( fd, recvbuf, recvlen );
  if ( rstatus < recvlen ) {
    return SCMD_RECV_ERR;
  }

  sg = (struct sg_header *)recvbuf;

  if ( !(sg->sense_buffer[0]) )
    return SCMD_NORM_END;
  else if ( (sg->sense_buffer[0] & 0x007f) == 0x0070 )
    return SCMD_SENSE_ERR;
  else if ( (sg->sense_buffer[0] & 0x007f)== 0x0071 )
    return SCMD_DIFFERED_ERR;
  else if ( sg->result )
    return SCMD_SEND_ERR;
  else
    return SCMD_BADSENSE_ERR;
}
