/*
   scsitool.c: get SCSI device information

   1996 (C) by JR0BAK
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/../scsi/scsi.h>
#include <unistd.h>
#include "var.h"
#include "scsicmd.h"
#include "callcmd.h"
#include "inquiry.h"
#include "tkdrive.h"
#include "unitready.h"
#include "rwbuffer.h"
#include "readcap.h"
#include "randomseek.h"

static void usage ( void );
static void tkcommand ( void );

void usage ( void )
{
  fputs ( "scsitool: display infomation of SCSI device\n" \
	 "usage: scsitool\n" \
	 "               do proper mode ( script or X mode )\n" \
	 "       scsitool -t\n" \
	 "               terminal mode ( not yet implemented )\n" \
	 "       scsitool -s\n" \
	 "               script mode\n" \
	 "       scsitool -h | -help | --help\n" \
	 "               show this message\n", stderr );
}


main ( int argc, char *argv[] )
{
  int tkpid = 0;
  int page = 0;

  if ( argc > 1
      && ( strcmp ( argv[1], "-h" ) == 0 
      || strcmp ( argv[1], "-help" ) == 0 
      || strcmp ( argv[1] , "--help" ) == 0 ) ) {
    usage ();
    exit ( 0 );
  } else if ( ( ( argc > 1 ) && strcmp ( argv[1], "-t" ) == 0 ) ) {
    tkpid = tkfork( TERMINAL );
    tkcommand ();
    wait ( NULL );
    exit ( 0 );
  } else if ( ( ( argc > 1 ) && strcmp ( argv[1], "-s" ) == 0 ) 
	     || ( ( getenv ( "DISPLAY" ) == NULL ) ) &&  ( isatty ( 0 ) ) ) {
    term_init ();
    term_mode ();
    exit ( 0 );
  } else if ( getenv ( "DISPLAY" ) ) { /* X window is available */
    tkpid = tkfork( XWINDOW );
    tkcommand ();
    wait ( NULL );
    exit ( 0 );
  } else { /* from remote shell ?? */
    exit ( -1 );
  }
}


void tkcommand ( void )
{
  char buf[1024];

  for ( ;; ) {
    gets ( buf );
    if ( strstr ( buf, "INQUIRY" ) ) {
      inquiry ( strchr ( buf, '/' ) );

    } else if ( strstr ( buf, "P-DEFECT" ) ) {
      read_defect ( 'p', buf[10], strchr ( buf, '/' ) );

    } else if ( strstr ( buf, "G-DEFECT" ) ) {
      read_defect ( 'g', buf[10], strchr ( buf, '/' ) );

    } else if ( strstr ( buf, "MSENSE-LIST" ) ) {
      mode_page ( strchr ( buf, '/' ) );

    } else if ( strstr ( buf, "MSENSE-PAGE" ) ) {
      /* "!MSENSE-PAGE 000 000 001 /dev/sga\n" */
      mode_sense ( atoi ( buf + 12 ), atoi ( buf + 16 ), strchr ( buf, '/' ),
		  atoi ( buf + 20 ) );

    } else if ( strstr ( buf, "FORMAT" ) ) {
      format_unit ( strchr ( buf, '/' ) );

    } else if ( strstr ( buf, "READCAP" ) ) {
      read_capacity ( strchr ( buf, '/' ) );

    } else if ( strstr ( buf, "RWBUFFER_INFO" ) ) {
      rwbuffer_info ( strchr ( buf, '/' ) );

    } else if ( strstr ( buf, "RBUFFER_BENCH" ) ) {
      /* "!RBUFFER_BENCH 001 /dev/sga\n" */
      rwbuffer_bench ( strchr ( buf, '/' ), 'R', atoi ( strchr ( buf, ' ' ) ) );

    } else if ( strstr ( buf, "WBUFFER_BENCH" ) ) {
      /* "!WBUFFER_BENCH 001 /dev/sga\n" */
      rwbuffer_bench ( strchr ( buf, '/' ), 'W', atoi ( strchr ( buf, ' ' ) ) );

    } else if ( strstr ( buf, "UNITBENCH" ) ) {
      /* "!UNITREADY 001 /dev/sga\n" */
      unitbench ( strchr ( buf, '/' ), atoi ( strchr ( buf, ' ' ) ) );

    } else if ( strstr ( buf, "UNITREADY" ) ) {
      unitready ( strchr ( buf, '/' ) );

    } else if ( strstr ( buf, "SEEKBENCH" ) ) {
      /* "!SEEKBENCH 001 /dev/sga\n" */
      randomseek ( strchr ( buf, '/' ), atoi ( strchr ( buf, ' ' ) ) );

    } else if ( strstr ( buf, "READBENCH" ) ) {
      /* "!SEEKBENCH 001 /dev/sga\n" */
      randomread ( strchr ( buf, '/' ), atoi ( strchr ( buf, ' ' ) ) );

    } else if ( strstr ( buf, "VERIFY" ) ) {
      verify ( strchr ( buf, '/' ) );

    } else if ( strstr ( buf, "MAKEERR" ) ) {
      makeerr ( strchr ( buf, '/' ), atoi ( strchr ( buf, ' ' ) + 1 ) );

    } else if ( strstr ( buf, "REASSIGN" ) ) {
      reassign ( strchr ( buf, '/' ), atoi ( strchr ( buf, ' ' ) + 1 ) );

    } else if ( strstr ( buf, "EXIT" ) ) {
      break;
    }
  }
}
