/*
   tkdrive.c: interface with wish
*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "tkdrive.h"
#include "../tk/ifscsitool.h"
#include "../tcl/tclscsitool.h"

#define FNAME "/tmp/ifscsitool.%d"
#define TTYIN  3
#define TTYOUT 4

static void err_busy ( void );
static void err_bug( void );
static void signal_handler ( int );
static int tkdrive_wish_pid = -1;

pid_t tkfork ( int prog )
{
  int p_ctotk[2], p_tktoc[2];
  int pid, i;
  char fname[256], cmdline[256];
  FILE *stream;

  /* expand wish script */
  sprintf ( fname, FNAME, (int)(getpid()) );
  sprintf ( cmdline, ZCAT "> %s", fname );
  stream = popen ( cmdline, "w" );
  if ( prog == XWINDOW )
    fwrite ( IFSCSITOOL, sizeof(char), IFSCSITOOL_SIZE, stream );
  else
    fwrite ( TCLSCSITOOL, sizeof(char), TCLSCSITOOL_SIZE, stream );
  pclose ( stream );
  chmod ( fname, S_IXUSR );

  /* add a `suicide file' command to the script */
  /* sprintf ( cmdline, "exec /bin/rm -f %s\n", fname );
  stream = fopen ( fname, "a" );
  fputs ( cmdline, stream );
  fclose ( stream ); */

  /* make a pipe pair */
  switch ( pipe ( p_ctotk ) ) {
  case 0: /* normally end, nothing to do */
    break;
  case EMFILE:
  case ENFILE: err_busy(); break;
  case EFAULT:
  default: err_bug(); break;
  }

  switch ( pipe ( p_tktoc ) ) {
  case 0: /* normally end, nothing to do */
    break;
  case EMFILE:
  case ENFILE: 
  case EFAULT:
  default: err_bug(); break;
  }

  /* run an interface process,
     and disconnect/reconnect streams */
/*  close ( TTYIN );  close ( TTYOUT );
  if ( dup2 ( 0, TTYIN ) != TTYIN ) err_busy();
  if ( dup2 ( 1, TTYOUT ) != TTYOUT ) err_busy();*/
  close ( 0 ); close ( 1 );

  switch ( pid = fork () ) {
  case 0: /* I'm a child process */
    if ( dup2 ( p_ctotk[0], 0 ) != 0 ) err_busy();
    if ( dup2 ( p_tktoc[1], 1 ) != 1 ) err_busy();
    close ( p_ctotk[1] ); close ( p_tktoc[0] );
    execlp ( fname, fname, fname, NULL );
    /* if execlp is succeed, below line cannot execute... */
    fputs ( "can't fork...\n", stderr );
    puts ( "!EXIT" );
    exit ( -1 );
    break;
  case EAGAIN: /* system is too busy to fork process ... */
    err_busy(); break;
  default: /* I'm a parent process. */
    for ( i = 1; i < 32; ++i )
      signal ( i, &signal_handler );
    tkdrive_wish_pid = pid;
    setbuf ( stdout, NULL );
    setbuf ( stdin, NULL );
    setbuf ( stderr, NULL );
    if ( dup2 ( p_tktoc[0], 0 ) != 0 ) err_busy();
    if ( dup2 ( p_ctotk[1], 1 ) != 1 ) err_busy();
    close ( p_ctotk[0] ); close ( p_tktoc[1] );
    break;
  }
  return pid;
}



static void err_busy ( void )
{
  fprintf ( stderr, "*** System is too busy.\n" );
  if ( tkdrive_wish_pid > 0 ) {
    kill ( tkdrive_wish_pid, SIGINT );
    wait ( tkdrive_wish_pid );
  }
  exit ( -1 );
}



static void err_bug( void )
{
  fprintf ( stderr, "*** BUG! BUG! BUG!\n" );
  if ( tkdrive_wish_pid > 0 ) {
    kill ( tkdrive_wish_pid, SIGINT );
    wait ( tkdrive_wish_pid );
  }
  exit ( -1 );
}



static void signal_handler ( int signum )
{
  if ( signum == SIGCONT )
    return;

  if ( tkdrive_wish_pid > 0 ) {
    kill ( tkdrive_wish_pid, SIGINT );
    wait ( tkdrive_wish_pid );
  }
  signum = (signum == 17) ? 0 : signum;
  exit ( signum );
}
